/**
 * Copyright (C) 2003 Michel Santos
 * Licensed under the Academic Free License version 3.0
 * Please see the LICENSE for more information.
 */

package cern.colt.matrix;


import cern.colt.matrix.*;


/**
 * A utility class with some functionality added on to the original {@link DoubleFactory2D}
 *
 * @see #appendColumns(DoubleMatrix2D, DoubleMatrix2D)
 * @see #random(int, int, double)
 */
public class DoubleFactory2D2 extends DoubleFactory2D {

  /**
   * A factory producing dense matrices.
   */
  public static final DoubleFactory2D2 dense  = new DoubleFactory2D2();
  
  /**
   * A factory producing sparse hash matrices.
   */
  public static final DoubleFactory2D2 sparse = new DoubleFactory2D2();
  
  /**
   * A factory producing sparse row compressed matrices.
   */
  public static final DoubleFactory2D2 rowCompressed = new DoubleFactory2D2();
  
  /**
   * Makes this class non instantiable, but still let's others inherit from it.
   */
  protected DoubleFactory2D2() {}

  /**
   * Append the columns from two matrices to form a new matrix
   *
   * @param A The first matrix
   * @param B The second matrix
   *
   * @return  The new combined matrix
   */
  public DoubleMatrix2D appendColumns(DoubleMatrix2D A, DoubleMatrix1D x) {
    // force both to have maximal shared number of rows.
    if (x.size() > A.rows()) x = x.viewPart(0,A.rows());
    else if (x.size() < A.rows()) A = A.viewPart(0,0,x.size(),A.columns());
    
    // concatenate
    int ac = A.columns();
    int bc = 1;
    int r = A.rows();
    DoubleMatrix2D matrix = make(r,ac+bc);
    matrix.viewPart(0,0,r,ac).assign(A);
    matrix.viewColumn(ac).assign(x);
    return matrix;
  }
  
  /**
   * Constructs a matrix with uniformly distributed values in <tt>(-range, +range)</tt> (exclusive).
   * Random values constructed using the {@link cern.jet.random.engine.MersenneTwister}
   *
   * @param rows  The number of rows desired for the matrix
   * @param columns  The number of columns desired for the matrix
   * @param range The range around 0
   *
   * @return  The matrix containing random values
   */
  public DoubleMatrix2D random(int rows, int columns, double range) {    
    return make(rows,columns).assign(new RandomCustom(range));
  }
  
  /**
   * A random number generator producing uniformly distributed values between -range and + range (exclusive)
   */
  protected class RandomCustom extends cern.jet.random.engine.MersenneTwister {
    private double dRange;
    public RandomCustom(double rangeOfValues) {
      super(new java.util.Date());
      this.dRange = rangeOfValues;
    }
    
    public double apply(double dummy) {
      return (super.apply(dummy) - 0.5) * dRange * 2;
    }
  }
}
