/**
 * Copyright (C) 2003 Michel Santos
 * Licensed under the Academic Free License version 3.0
 * Please see the LICENSE for more information.
 */

package cern.colt.matrix.linalg;


import cern.colt.matrix.*;


/**
 * Linear algebraic matrix operations operating on {@link DoubleMatrix2D}.
 * This extension of the original {@link Algebra} class aims to accomplish the following things:
 * <OL>
 *  <LI>Concentrates some functionality that is available in other parts of this package</LI>
 *  <LI>Adds some ease-of-use functionality</LI>
 *  <LI>Corrects a bug in the original {@link #norm2(DoubleMatrix1D)}</LI>
 * </OL>
 *
 * @author  Michel Santos
 * @since ?.?
 */

public class Algebra2 extends cern.colt.matrix.linalg.Algebra {
  
  /**
   * A default Algebra object; has {@link Property#DEFAULT} attached for tolerance.
   * Allows ommiting to construct an Algebra object time and again.
   *
   * Note that this Algebra object is immutable.
   * Any attempt to assign a new Property object to it (via method <tt>setProperty</tt>), or to alter the tolerance of its property object (via <tt>property().setTolerance(...)</tt>) will throw an exception.
   */
  public static final Algebra2 DEFAULT2;
  
  /**
   * A default Algebra object; has {@link Property#ZERO} attached for tolerance.
   * Allows ommiting to construct an Algebra object time and again.
   *
   * Note that this Algebra object is immutable.
   * Any attempt to assign a new Property object to it (via method <tt>setProperty</tt>), or to alter the tolerance of its property object (via <tt>property().setTolerance(...)</tt>) will throw an exception.
   */
  public static final Algebra2 ZERO2;
  
  
  static {
    // don't use new Algebra(Property.DEFAULT.tolerance()), because then property object would be mutable.
    DEFAULT2 = new Algebra2();
    DEFAULT2.property = Property.DEFAULT; // immutable property object
    
    ZERO2 = new Algebra2();
    ZERO2.property = Property.ZERO; // immutable property object
  }
  
  
  /**
   * Constructs a new instance with an equality tolerance given by <tt>Property.DEFAULT.tolerance()</tt>.
   */
  public Algebra2() {
    this(Property.DEFAULT.tolerance());
  }
  /**
   * Constructs a new instance with the given equality tolerance.
   * @param tolerance the tolerance to be used for equality operations.
   */
  public Algebra2(double tolerance) {
    setProperty(new Property(tolerance));
  }
  
  /**
   * Modifies the matrix to be an upper trapezoidal matrix.
   *
   * @return <tt>A</tt> (for convenience only).
   * @see #triangulateLower(DoubleMatrix2D)
   */
  protected DoubleMatrix2D trapezoidalUpper(DoubleMatrix2D A) {
    int rows = A.rows();
    int columns = A.columns();
    for (int r = rows; --r >= 0; ) {
      for (int c = columns; --c >= 0; ) {
        if (r > c) A.setQuick(r,c, 0);
      }
    }
    return A;
  }

  private class FuncMultiply implements cern.colt.function.DoubleFunction {
    private double scalar = 1;
    FuncMultiply() {}
    public void setScalar(double newScalar) {this.scalar = newScalar;}
    public double apply(double cellValue) {
      return cellValue * scalar;
    }
  }
  private FuncMultiply funcMult = new FuncMultiply();
//  public DoubleMatrix1D multEq(DoubleMatrix1D other, double scalar) {
//    funcMult.setScalar(scalar);
//    return other.assign(funcMult);
//  }
  public DoubleMatrix1D multEq(DoubleMatrix1D other, double scalar) {
    int iSize = other.size();
    for (int i = 0; i < iSize; i++) {
      double dValue = other.getQuick(i);
      dValue *= scalar;
      other.setQuick(i, dValue);
    }
    return other;
  }
  public DoubleMatrix1D mult(DoubleMatrix1D other, double scalar) {
    DoubleMatrix1D result = other.copy();
    return multEq(result, scalar);
  }
  public DoubleMatrix1D divEq(DoubleMatrix1D other, double divisor) {
//    funcMult.setScalar(1 / divisor);
//    return other.assign(funcMult);
    int iSize = other.size();
    for (int i = 0; i < iSize; i++) {
      double dValue = other.getQuick(i);
      dValue /= divisor;
      other.setQuick(i, dValue);
    }
    return other;    
  }
  public DoubleMatrix1D div(DoubleMatrix1D other, double divisor) {
    DoubleMatrix1D result = other.copy();
//    funcMult.setScalar(1.0 / divisor);
//    return result.assign(funcMult);
    return divEq(result, divisor);
  }
  
  
  
  
  private class FuncAdd implements cern.colt.function.DoubleFunction {
    private double extra = 0;
    FuncAdd() {}
    public void setExtra(double newExtra) {this.extra = newExtra;}
    public double apply(double cellValue) {
      return cellValue + extra;
    }
  }
  private FuncAdd funcAdd = new FuncAdd();
  /** @todo Test whether a "faster" version of "plusEq" can be made by
   * looping through the contents */
  public DoubleMatrix1D plusEq(DoubleMatrix1D other, double value) {
    funcAdd.setExtra(value);
    return other.assign(funcAdd);
  }
  public DoubleMatrix1D plus(DoubleMatrix1D other, double value) {
    DoubleMatrix1D result = other.copy();
    funcAdd.setExtra(value);
    return result.assign(funcAdd);
  }
  /** @todo Test whether a "faster" version of "minusEq" can be made by
   * looping through the contents */
  public DoubleMatrix1D minusEq(DoubleMatrix1D other, double value) {
    funcAdd.setExtra(-value);
    return other.assign(funcAdd);
  }
  public DoubleMatrix1D minus(DoubleMatrix1D other, double value) {
    DoubleMatrix1D result = other.copy();
    funcAdd.setExtra(-value);
    return result.assign(funcAdd);
  }
  
  public DoubleMatrix1D plusEq(DoubleMatrix1D toThat, DoubleMatrix1D addThis) {
    toThat.assign(addThis, cern.jet.math.Functions.plus);
    return toThat;
  }
  public DoubleMatrix1D plus(DoubleMatrix1D a, DoubleMatrix1D b) {
    DoubleMatrix1D result = a.copy();
    result.assign(b, cern.jet.math.Functions.plus);
    return result;
  }
  public DoubleMatrix1D minusEq(DoubleMatrix1D toThat, DoubleMatrix1D subtractThis) {
    toThat.assign(subtractThis, cern.jet.math.Functions.minus);
    return toThat;
  }
  public DoubleMatrix1D minus(DoubleMatrix1D a, DoubleMatrix1D b) {
    DoubleMatrix1D result = a.copy();
    result.assign(b, cern.jet.math.Functions.minus);
    return result;
  }

  
  /**
   * Add the contents of two matrices together
   */
  public DoubleMatrix2D plus(DoubleMatrix2D a, DoubleMatrix2D b) {
    DoubleMatrix2D result = a.copy();    
    return plusEq(result, b);
  }

  /**
   * Add the contents of two matrices together
   */
  public DoubleMatrix2D plusEq(DoubleMatrix2D a, DoubleMatrix2D b) {
    DoubleMatrix2D result = a.assign(b, cern.jet.math.Functions.plus);
    return result;
  }

  
  /**
   * Multiply the contents of a matrix by a scalar
   */
  public DoubleMatrix2D multEq(DoubleMatrix2D matrix, double scalar) {
    int iRows = matrix.rows();
    int iCols = matrix.columns();
    
    for (int iRow = 0; iRow < iRows; iRow++) {
      for (int iCol = 0; iCol < iCols; iCol++) {
        double dValue = matrix.getQuick(iRow, iCol);
        dValue *= scalar;
        matrix.setQuick(iRow, iCol, dValue);
      }
    }
    
    return matrix;
  }
  

  
  
  /**
   * Find the maximum of the values of the vector
   */
  public double max(DoubleMatrix1D x) {
    //if (x.size()==0) return 0;
    //return x.aggregate(cern.jet.math.Functions.plus,cern.jet.math.Functions.abs);
    double max = 0;
    int iSize = x.size();
    for (int i = iSize; --i >= 0; ) {
      max = Math.max(max, x.getQuick(i));
    }
    return max;
  }
  
  /**
   * Find the maximum of the absolute values of the vector
   */
  public double maxAbs(DoubleMatrix1D x) {
    //if (x.size()==0) return 0;
    //return x.aggregate(cern.jet.math.Functions.plus,cern.jet.math.Functions.abs);
    double max = 0;
    int iSize = x.size();
    for (int i = iSize; --i >= 0; ) {
      max = Math.max(max, Math.abs(x.getQuick(i)));
    }
    return max;
  }  
  
  /**
   * Find the minimum of the values of the vector
   */
  public double min(DoubleMatrix1D x) {
    double dMin = Double.POSITIVE_INFINITY;
    int iSize = x.size();
    for (int i = iSize; --i >= 0; ) {
      dMin = Math.min(dMin, x.getQuick(i));
    }
    return dMin;
  }
  
  /**
   * Find the minimum of the absolute values of the vector
   */
  public double minAbs(DoubleMatrix1D x) {
    double dMin = Double.POSITIVE_INFINITY;
    int iSize = x.size();
    for (int i = iSize; --i >= 0; ) {
      dMin = Math.min(dMin, Math.abs(x.getQuick(i)));
    }
    return dMin;
  }

  
  /**
   * Swap two columns of a matrix
   */
  public void permuteColumns(DoubleMatrix2D matrix, int firstColumn, int secondColumn) {
    // Safety-check
    if (matrix == null) {throw new IllegalArgumentException("A matrix was not provided");}
    
    int iCols = matrix.columns();
    // Safety-check
    if (firstColumn <0 || secondColumn < 0 || firstColumn >= iCols || secondColumn >= iCols) {
      throw new IllegalArgumentException("Valid column numbers were not provided");      
    }
    if (firstColumn == secondColumn) {
      // Do nothing
      return;
    }
    
    
    // Create standard array of column numbers
    int[] iAryColumnIndices = new int[iCols];
    for (int i = 0; i < iCols; i++) {
      iAryColumnIndices[i] = i;      
    }
    
    // Switch the index of the columns specified
    iAryColumnIndices[firstColumn] = secondColumn;
    iAryColumnIndices[secondColumn] = firstColumn;
    
    int[] iArySomehowThisIsUsefule = new int[iCols];
    permuteColumns(matrix, iAryColumnIndices, iArySomehowThisIsUsefule);
  }
  
  /**
   * Returns the two-norm (aka <i>euclidean norm</i>) of vector <tt>x</tt>.  
   * Square root of the sum of the square of the elements.
   * <br>
   * Overrides the incorrect implementation in the original Algebra object.  
   * See the L2-norm definition at MathWorld {@link http://mathworld.wolfram.com/L2-Norm.html}
   */
  public double norm2(DoubleMatrix1D x) {
    return Math.sqrt(mult(x,x));
  }

  private DoubleFactory2D factory2D = DoubleFactory2D.dense;
  private DoubleFactory1D factory1D = DoubleFactory1D.dense;
   /**
   * Solve a matrix equation of the form A*x = b for the value of x
   *
   * @param A The A matrix
   * @param b The b vector
   *
   * @return The vector solution in the form of DoubleMatrix1D
   */
 public DoubleMatrix1D solve(DoubleMatrix2D A, DoubleMatrix1D b) {
    if (A.rows() >= A.columns()) {
      DoubleMatrix2D matSolution = solve(A, factory2D.make(b.toArray(), b.size()));
      return matSolution.viewColumn(0);
      
    } else {
      return QRDecomposition2.solveRankDeficient(A, b);
    }
  }
  

  /**
   * Make a vector full of zeros
   *
   * @param size  Size of vector
   * @return  A vector in the form of a DoubleMatrix1D
   */
  public DoubleMatrix1D zeros(int size) {
    return factory1D.make(size, 0);
  }
  
  public DoubleMatrix2D multRowEq(DoubleMatrix2D matrix, int rowNbr, double scalar) {
    // Safety-check
    if (matrix == null) {
      throw new IllegalArgumentException("A valid matrix must be provided");      
    }
    
    // Safety-check
    if (rowNbr < 0 || rowNbr >= matrix.rows()) {
      throw new IndexOutOfBoundsException("The specified row number (" + rowNbr + ") is out of bounds.");
    }
    
    int iCols = matrix.columns();
    
    for (int iCol = 0; iCol < iCols; iCol++) {
      double dValue = matrix.getQuick(rowNbr, iCol);
      dValue *= scalar;
      matrix.setQuick(rowNbr, iCol, dValue);
    }
    
    return matrix;
    
  }
  
  
  public DoubleMatrix2D multColumnEq(DoubleMatrix2D matrix, int columnNbr, double scalar) {
    // Safety-check
    if (matrix == null) {
      throw new IllegalArgumentException("A valid matrix must be provided");      
    }
    
    // Safety-check
    if (columnNbr < 0 || columnNbr >= matrix.columns()) {
      throw new IndexOutOfBoundsException("The specified row number (" + columnNbr + ") is out of bounds.");
    }
    
    int iRows = matrix.rows();    
    for (int iRow = 0; iRow < iRows; iRow++) {
      double dValue = matrix.getQuick(iRow, columnNbr);
      dValue *= scalar;
      matrix.setQuick(iRow, columnNbr, dValue);
    }
    
    return matrix;
    
  }

  
}