/**
 * Copyright (C) 2003 Michel Santos
 * Licensed under the Academic Free License version 3.0
 * Please see the LICENSE for more information.
 *
 * Created on August 7, 2003, 5:28 PM
 */

package cern.colt.matrix.linalg;


import cern.colt.matrix.*;


/**
 * Performs consistency and redundancy checks on linear systems of equations (A * x = b)
 *
 * Makes extensive use of {@link QRDecomposition2}
 *
 * @author  Michel Santos
 * @since ?.?
 */
public class Checker {
  
  /** 
   * Creates a new instance of Checker
   */
  public Checker(DoubleMatrix2D A) {
    // Safety-check
    if (A == null) {throw new IllegalArgumentException("A valid 2D matrix must be provided");}
    int iRows = A.rows();
    DoubleMatrix1D vecB = cern.colt.matrix.DoubleFactory1D.sparse.make(iRows, 0);
    
    init(A, vecB);
  }  
  
  private DoubleMatrix2D matA = null;
  private DoubleMatrix1D vecB = null;  
  public Checker(DoubleMatrix2D A, DoubleMatrix1D b) {
    init(A, b);
  }

  /**
   * Initialize everything
   *
   * @param A
   * @param b
   */
  private void init(DoubleMatrix2D A, DoubleMatrix1D b) {
    dimensionCheck(A, b);
    
    // Protect the input matrices
    A = A.copy();
    b = b.copy();
    
    scaleEquality(A, b);
    
    int iVariables = A.columns();
    dTolerance = 100 * iVariables * 2e-16;
    
    matA = A;
    vecB = b;
  }
  
  /**
   * Check the dimensions
   */
  private static void dimensionCheck(DoubleMatrix2D A, DoubleMatrix1D b) {
    // Safety-check
    if (A == null) {throw new IllegalArgumentException("A valid 2D matrix must be provided");}
    if (b == null) {throw new IllegalArgumentException("A valid 1D matrix must be provided");}
    int iRows = A.rows();
    int iCols = A.columns();
    if (iRows != b.size()) {throw new IllegalArgumentException("Both A and b must have the same number of rows");}
  }

  protected static cern.colt.matrix.linalg.Algebra2 LA = new cern.colt.matrix.linalg.Algebra2();
  
  /**
   * Scaling A * x = b by the norm of each row of A
   */
  public static void scaleEquality(DoubleMatrix2D A, DoubleMatrix1D b) {
    dimensionCheck(A, b);
    int iRows = A.rows();
    
    // Loop through each row
    for (int iRow = 0; iRow < iRows; iRow++) {
      // Determine norm of each row
      DoubleMatrix1D vecRow = A.viewRow(iRow);
      double dNorm = LA.norm2(vecRow);
      
      // Scale each row by the norm
      LA.divEq(vecRow, dNorm);
      b.setQuick(iRow,  b.getQuick(iRow) / dNorm);
    }
  }
  
  /**
   * Determines whether the system of equations A * x = b are consistent
   */
  public static boolean isConsistent(DoubleMatrix2D A, DoubleMatrix1D b) {
    int[] iAryInconsistentEquations = listInconsistentEquations(A, b);
    return iAryInconsistentEquations.length == 0;
  }
  
  /**
   * Lists the equations within the system of equations A * x = b that are inconsistent.
   * 
   * @return  An array containing the indices of the equations.  The indices are zero-based.
   */
  public static int[] listInconsistentEquations(DoubleMatrix2D A, DoubleMatrix1D b) {
    dimensionCheck(A, b);    
    int iRows = A.rows();
    int iCols = A.columns();
    int iVariables = iCols;
    int iEquations = iRows;
    
    // Determine the maximum number of variables that can be solved for
    // by determining the span of variables than can be reached
    // This is done by performing a QRDecomposition on the equations
    // Given a rows x cols A matrix
    //  Q will be rows x rows, R will be rows x cols, Pivots will be cols x cols
    //  so that A * Pivots = Q * R where Pivots is an identity matrix with permuted columns
    //  such the Q's diagonal elements are in absolute magnitude, decreasing order
    QRDecomposition2 qr = new QRDecomposition2(A, true);
    double dTOL = 100 * iVariables * 2e-16;
    qr.setTolerance(dTOL);
    int iIndeterminateVariables = qr.getDependentColumns().length;    
    int iExtraEqns = Math.max(iEquations - (iVariables - iIndeterminateVariables), 0);

    
    // System.out.println("A: " + A);
    // System.out.println("b: " + b);
    // System.out.println("Q*R*I': " + LA.mult(LA.mult(qr.getQ(), qr.getR()), qr.getPivots().viewDice()) );

    
    if (iExtraEqns == 0) {
      //// System.out.println("qr: " + qr);
      //// System.out.println("qr's Pivots: " + qr.getPivots());
      // System.out.println("No extra equations");
      return new int[0];
      
    } else {    
      // A*x = b; A*E = Q*R where E is the matrix of column permutations
      // A = Q*R*inv(E) = Q*R*E'
      // A*x = Q*R*E'*x = b
      // R*E*x = inv(Q)*b = Q'*b
      // For consistency, the redundant equations, after they have been QRed,
      // should have 0 on the right-hand-side of the equations
      // These will be the "lower" equations
      DoubleMatrix2D matQ = qr.getQ();
      int[] iAryExtraEqns = new int[iExtraEqns];
      int iValidVariables = qr.getIndependentColumns().length;
      int iValidEquations = iValidVariables;      
      for (int iVar = 0; iVar < iExtraEqns; iVar++) {
        iAryExtraEqns[iVar] = iValidEquations + iVar;
      }
      // System.out.println("iValidEquations: " + iValidEquations);
      // System.out.println("iValidVariables: " + iValidVariables);
      // System.out.println("iExtraEqns: " + iExtraEqns);
      // System.out.println("iIndeterminateVariables: " + iIndeterminateVariables);
      // print(iAryExtraEqns);
      // System.out.println("qr: " + qr);
      matQ = matQ.viewSelection(null, iAryExtraEqns); // Grab the "right" columns
      matQ = matQ.viewDice(); // So the "right" columns are now the "lower" rows
      
      DoubleMatrix1D vecRHS = LA.mult(matQ, b); // Right-hand side
      
      // Check whether any elements of the RHS are "larger" than zero
      // If so, then inconsistent
      int iSize = vecRHS.size();  // Should be equal to the number of "extra" equations
      // System.out.println("Residual: " + vecRHS);
      org.xspace.util.IntegerList list = new org.xspace.util.IntegerList(iSize);
      for (int i = 0; i < iSize; i++) {
        double dValue = vecRHS.getQuick(i);
        // if (Math.abs(dValue) > dTOL) {return false;}
        if (Math.abs(dValue) > dTOL) {list.add(i);}
      }
      // return true;
      return list.toArray();
    }
  }
  
  /**
   * Determine whether any of the rows of the matrix A are redundant
   */
  public static boolean isRedundant(DoubleMatrix2D A) {
    // return (qr.getDependentColumns().length > 0);
    return (listRedundantEquations(A).length > 0);
  }

  /**
   * Lists the rows of the matrix A that are redundant
   * 
   * @return  An array containing the indices of the equations.  The indices are zero-based.
   */
  public static int[] listRedundantEquations(DoubleMatrix2D A) {
    int iRows = A.rows();
    int iCols = A.columns();
    int iVariables = iCols;
    int iEquations = iRows;
    
    // Perform QR Decomposition on the equations (rows) of A
    // by transposing the matrix and then analyzing the columns
    QRDecomposition2 qr = new QRDecomposition2(A.viewDice(), true);
    double dTOL = 100 * iVariables * 2e-16;
    qr.setTolerance(dTOL);
        
    return qr.getDependentColumns();
  }

  /**
   * Lists the rows of the matrix A that are unique/not redundant
   * 
   * @return  An array containing the indices of the equations.  The indices are zero-based.
   */
  public static int[] listUniqueEquations(DoubleMatrix2D A) {
    int iRows = A.rows();
    int iCols = A.columns();
    int iVariables = iCols;
    int iEquations = iRows;
    
    // Perform QR Decomposition on the equations (rows) of A
    // by transposing the matrix and then analyzing the columns
    QRDecomposition2 qr = new QRDecomposition2(A.viewDice(), true);
    double dTOL = 100 * iVariables * 2e-16;
    qr.setTolerance(dTOL);
        
    return qr.getIndependentColumns();
  }
  
//  public static boolean isRedundant(DoubleMatrix2D A, DoubleMatrix1D b) {
//    dimensionCheck(A, b);
//    
//    int iRows = A.rows();
//    int iCols = A.columns();
//    int iVariables = iCols;
//    int iEquations = iRows;
//    
//    // Perform QR Decomposition on the equations (rows) of A
//    // by transposing the matrix and then analyzing the columns
//    QRDecomposition4 qr = new QRDecomposition4(A.viewDice(), true);
//    double dTOL = 100 * iVariables * 2e-16;
//    qr.setTolerance(dTOL);
//        
////    // Keep only the independent columns (equations) and the be vector
////    // and transpose the columns back to rows
////    int[] iIndependentEqns = qr.getIndependentColumns();
////    A = A.viewSelection(iIndependentEqns, null);
////    b = b.viewSelection(iIndependentEqns);    
////    // System.out.println("A: " + A);
////    // System.out.println("b: " + b);
//    
//    return (qr.getDependentColumns().length > 0);
//  }
  
  
  
  private boolean bIsConsistent = false;
  private boolean bConsistencyAlreadyChecked = false;
  public boolean isConsistent() {
    // Safety-check
    if (matA == null || vecB == null) {throw new IllegalStateException("No valid matrix and vector were provided");}   
    
    if (!bConsistencyAlreadyChecked) {
      bConsistencyAlreadyChecked = true;
      bIsConsistent = isConsistent(matA, vecB);
    }
    
    return bIsConsistent;    
  }
  
  private QRDecomposition2 qrRedundancy = null;
  
  
  
  private boolean bRedundancyAlreadyChecked = false;
  private boolean bRedundant = false;
  private boolean bInconsistent = false;
  /**
   * Determines whether there are any redundant equations in the linear algebra problem, A * x = b
   *
   * @throws  InconsistentEquationsException  If the given equations are inconsistent
   */
  public boolean isRedundant() throws InconsistentEquationsException {
    if (bRedundancyAlreadyChecked) {
      if (bInconsistent) {
        throw new InconsistentEquationsException();
        
      } else {
        return bRedundant;
      }     
    }
    
    
    // Safety-check
    if (matA == null || vecB == null) {throw new IllegalStateException("No valid matrix and vector were provided");}
    if (!isConsistent(matA, vecB)) {
      bInconsistent = true;
      throw new InconsistentEquationsException();
    } else {
      bInconsistent = false;
    }
    
    
    if (qrRedundancy == null) {
      int iRows = matA.rows();
      int iCols = matA.columns();
      int iVariables = iCols;
      int iEquations = iRows;

      // Perform QR Decomposition on the equations (rows) of A
      // by transposing the matrix and then analyzing the columns
      qrRedundancy = new QRDecomposition2(matA.viewDice(), true);
      qrRedundancy.setTolerance(dTolerance);
    }
    
    bRedundant = (qrRedundancy.getDependentColumns().length > 0);
    bRedundancyAlreadyChecked = true;
    return bRedundant;
  }
  
  
  /**
   * Lists the rows of the matrix A that are redundant
   * 
   * @return  An array containing the indices of the equations.  The indices are zero-based.
   */
  public int[] listRedundantEquations() throws InconsistentEquationsException {
    isRedundant();
    
    return qrRedundancy.getDependentColumns();
  }

  /**
   * Lists the rows of the matrix A that are unique/not redundant
   * 
   * @return  An array containing the indices of the equations.  The indices are zero-based.
   */
  public int[] listUniqueEquations() throws InconsistentEquationsException {
    isRedundant();
    
    return qrRedundancy.getIndependentColumns();
  }
  
  
  private double dTolerance = 1e-6;
  public void setTolerance(double tolerance) {
    // Safety-check
    if (tolerance <= 0) {throw new IllegalArgumentException("A positive values tolerance is required");}
    
    dTolerance = tolerance;
  }
  
  private static void print(int[] array) {
    for (int i = 0; i < array.length; i++) {
      System.out.println("Index #" + i + " = " + array[i]);
    }
  }  
  
}
