/*
 * Copyright (C) 2003 Michel Santos
 * Licensed under the Academic Free License version 3.0
 * Please see the LICENSE for more information.
 */

package cern.colt.matrix.linalg;


import cern.colt.matrix.*;
import cern.jet.math.Functions;
import org.xspace.util.IntegerList;


/**
 * For an <tt>m x n</tt> matrix <tt>A</tt> with <tt>m >= n</tt>, the QR decomposition is an <tt>m x n</tt>
 * orthogonal matrix <tt>Q</tt> and an <tt>n x n</tt> upper triangular matrix <tt>R</tt> so that
 * <tt>A = Q*R</tt>.
 * <br>
 * The QR decompostion always exists, even if the matrix does not have
 * full rank, so the constructor will never fail.  The primary use of the
 * QR decomposition is in the least squares solution of nonsquare systems
 * of simultaneous linear equations.  This will fail if <tt>isFullRank()</tt>
 * returns <tt>false</tt>.
 * <br>
 * This implementation uses the Householder QR factorization procedure.
 * A good description of this is available in Gene H. Golub's and Charles F. Van Loan's
 * book entitled <i>Matrix Computations</i>.
 * Furthermore, this implementation is different from the original QRDecomposition in Colt
 * by being capable of identifying for the user the dependent and independent columns of a matrix.
 * <br>
 * @author  Michel Santos
 * @since ?.?
 */
public class QRDecomposition2 implements java.io.Serializable {  
  static final long serialVersionUID = 1020;
  /**
   * Internal storage of decomposition.<br>
   * The upper triangle contains the R matrix.<br>
   * Beneath the upper triangle contains the essential parts of the Householder vectors<br>
   *
   * @serial internal array storage.
   */
  protected DoubleMatrix2D QR;  
  
  /**
   * Internal storage of the <i>beta</i> multiplier 
   * (of the Householder vector; used in computing the Q matrix)
   */
  protected double[] dAryMultiplier;
//  public DoubleMatrix2D[] aryP;
  
  public int[] iAryIndices;
  
  /**
   * Row and column dimensions.
   * @serial column dimension.
   * @serial row dimension.
   */
  protected int iRows, iCols;
  
  private cern.colt.matrix.linalg.Algebra2 LA = new cern.colt.matrix.linalg.Algebra2();
  private DoubleFactory2D factory2D = DoubleFactory2D.dense;
  private cern.colt.matrix.DoubleFactory1D factory1D = cern.colt.matrix.DoubleFactory1D.dense;
  
  /**   
   * Constructs and returns a new QR decomposition object;  computed by Householder reflections;
   * The decomposed matrices can be retrieved via instance methods of the returned decomposition object.
   * @param A    A rectangular matrix.
   * @return     a decomposition object to access <tt>R</tt> and the Householder vectors <tt>H</tt>, and to compute <tt>Q</tt>.
   * @throws IllegalArgumentException if <tt>A.rows() < A.columns()</tt>.
   */
  public QRDecomposition2(DoubleMatrix2D A) {
    init(A);
  }  
  
  /**   
   * Constructs and returns a new QR decomposition object;  computed by Householder reflections;
   * The decomposed matrices can be retrieved via instance methods of the returned decomposition object.
   * If pivoting is used, then determining redundant columns is possible.
   * @param A    A rectangular matrix.
   * @param pivot Specify whether or not pivoting is allowed
   * @return     a decomposition object to access <tt>R</tt> and the Householder vectors <tt>H</tt>, and to compute <tt>Q</tt>.
   * @throws IllegalArgumentException if <tt>A.rows() < A.columns()</tt>.
   */
  public QRDecomposition2(DoubleMatrix2D A, boolean pivot) {
    if (pivot) {
      initPivot(A);
    } else {
      init(A);
    }
  }  

  /**
   * Keeps track of how many extra rows of zeros were added to make the matrix
   * square.  This is necessary for the algorithm implemented.
   * This is then re-used to trim down the output of this approach, as if the
   * added zeros had never been there.
   */
  protected int iExtraRows = 0;
  
  /**
   * Constructs and returns a new QR decomposition object;  computed by Householder reflections;
   * As described by Algorithm 5.2.1 (Householder QR) p.224
   * The decomposed matrices can be retrieved via instance methods of the returned decomposition object.
   * @param A    A rectangular matrix.
   * @return     a decomposition object to access <tt>R</tt> and the Householder vectors <tt>H</tt>, and to compute <tt>Q</tt>.
   * @throws IllegalArgumentException if <tt>A.rows() < A.columns()</tt>.
   */
  protected void init(final DoubleMatrix2D A) {
    // Safety-check
    if (A == null) {throw new IllegalArgumentException("A non-null matrix be provided for decomposition");}
    iRows = A.rows();
    iCols = A.columns();
    // System.out.println("iRows: " + iRows + "; iCols: " + iCols);
    
    // Initialize
    if (iRows >= iCols) {
      QR = A.copy();  
    } else {
      iExtraRows = iCols - iRows;      
      DoubleMatrix2D matExtraZeros = factory2D.make(iExtraRows, iCols);
      // QR = DoubleFactory2D2.dense.appendRows(A, matExtraZeros);
      QR = factory2D.appendRows(A, matExtraZeros);
      
      iRows = QR.rows();      
      iCols = QR.columns();
      // System.out.println("iRows: " + iRows + "; iCols: " + iCols);
    }
    dAryMultiplier = new double[Math.min(iRows, iCols)];
//    aryP = new DoubleMatrix2D[Math.min(iRows, iCols)];
    iAryIndices = new int[iCols];
    for (int i = 0; i < iCols; i++) {iAryIndices[i] = i;}

    
    // Loop through all columns
    QRDecomposition2.HouseholderVector house = new QRDecomposition2.HouseholderVector();
    for (int iCol = 0; iCol < iCols; iCol++) {
      // Determine the Householder vector
      int iWidth = iCols - iCol;
      int iHeight = iRows - iCol;
      DoubleMatrix1D vecColumn = QR.viewColumn(iCol).viewPart(iCol, iHeight);

      DoubleMatrix1D vecReflector = house.reset(vecColumn);
      double dMultiplier = house.getMultiplier();
      
      // Store the multiplier
      dAryMultiplier[iCol] = dMultiplier;
      
      // Upper-triangulate (store) the A matrix
      // A = (I - beta * v * v') * A
      // A(iCol:iRows, iCol:iCols) = (I_(m-j+1) - beta * v * v') * A(iCol:iRows, iCol:iCols)      
      DoubleMatrix2D matP = house.getReflector();
//      aryP[iCol] = matP;
      DoubleMatrix2D matQRSubset = QR.viewPart(iCol, iCol, iHeight, iWidth);
      DoubleMatrix2D matTemp = matP.zMult(matQRSubset, null);      
      matQRSubset.assign(matTemp);
      
      // Store Householder vector in the zeroed-out lower-triangle
      // Assuming that the householder vector's first element is equal to one (1)
      // Ignore that first element
      if (iCol < iRows) {
        // A(iCol + 1:iRows, iCol) = v(2:iRows - iCol + 1)
        for (int iOffset = 1; iOffset < iHeight; iOffset++) {
          QR.setQuick(iCol + iOffset, iCol, vecReflector.getQuick(iOffset));
        }
      }
    }
  }

  
  /**
   * Constructs and returns a new QR decomposition object.  
   * Computed by Householder reflections as described by Algorithm 5.4.1 (Householder QR) p.248
   * <br>
   * The decomposed matrices can be retrieved via instance methods of the returned decomposition object.
   * <br>
   * @param A    A rectangular matrix.
   * @return     a decomposition object to access <tt>R</tt> and the Householder vectors <tt>H</tt>, and to compute <tt>Q</tt>.
   * @throws IllegalArgumentException if <tt>A.rows() < A.columns()</tt>.
   */
  protected void initPivot(final DoubleMatrix2D A) {
    // Safety-check
    if (A == null) {throw new IllegalArgumentException("A non-null matrix be provided for decomposition");}
    iRows = A.rows();
    iCols = A.columns();

    // Initialize
    if (iRows >= iCols) {
      QR = A.copy();  
    } else {
      iExtraRows = iCols - iRows;
      DoubleMatrix2D matExtraZeros = factory2D.make(iExtraRows, iCols);
      // QR = DoubleFactory2D2.dense.appendRows(A, matExtraZeros);
      QR = factory2D.appendRows(A, matExtraZeros);

      iRows = QR.rows();      
      iCols = QR.columns();
    }
    dAryMultiplier = new double[Math.min(iRows, iCols)];
//    aryP = new DoubleMatrix2D[Math.min(iRows, iCols)];
    iAryIndices = new int[iCols];
    for (int i = 0; i < iCols; i++) {iAryIndices[i] = i;}
    
    
    // Compute the square of the norm of each column
    double[] dAryNormSquared = new double[iCols];
    DoubleMatrix1D vecNormSquared = factory1D.make(iCols);
    for (int iCol = 0; iCol < iCols; iCol++) {
      DoubleMatrix1D vecColumn = QR.viewColumn(iCol);
      double dNormSquared = LA.mult(vecColumn, vecColumn);
      dAryNormSquared[iCol] = dNormSquared;
    }

    
    //  Loop through the columns
    QRDecomposition2.HouseholderVector house = new QRDecomposition2.HouseholderVector();
    for (int iCol = 0; iCol < iCols; iCol++) {
//      // System.out.println();
//      // System.out.println();
//      // System.out.println("iCol: " + iCol);
//      // System.out.println("Permutations: "); print(iAryIndices);
//      print(dAryNormSquared);
//      // System.out.println("Phase 1: " + QR);
      
      // Find the largest norm with the lowest-numbered index (arbitrary but effective)
      int iLargestIndex = findLargestValue(dAryNormSquared, iCol);
      double dLargestValue = dAryNormSquared[iLargestIndex];      
      // If the largest value is 0 then break out of loop
      if (dLargestValue <= 0) {
      // if (dLargestValue <= TOL) {
        break;
      }
      
      
      // Swap the largest column with the current column of interest      
      swap(iAryIndices, iCol, iLargestIndex);
      LA.permuteColumns(QR, iCol, iLargestIndex);
      swap(dAryNormSquared, iCol, iLargestIndex);
//      // System.out.println();
//      // System.out.println("Phase 2 (Swap): " + QR);
      
      
      // Upper triangularize (as per regular QR Decomposition)
      // Determine the Householder vector
      int iWidth = iCols - iCol;
      int iHeight = iRows - iCol;
      DoubleMatrix1D vecColumn = QR.viewColumn(iCol).viewPart(iCol, iHeight);
      
      DoubleMatrix1D vecReflector = house.reset(vecColumn);
      double dMultiplier = house.getMultiplier();
      
      // Store the multiplier
      dAryMultiplier[iCol] = dMultiplier;
      
      // Upper-triangulate (store) the A matrix
      // A = (I - beta * v * v') * A
      // A(iCol:iRows, iCol:iCols) = (I_(m-j+1) - beta * v * v') * A(iCol:iRows, iCol:iCols)
      DoubleMatrix2D matP = house.getReflector();
//      aryP[iCol] = matP;
      DoubleMatrix2D matQRSubset = QR.viewPart(iCol, iCol, iHeight, iWidth);
      DoubleMatrix2D matTemp = matP.zMult(matQRSubset, null);      
      matQRSubset.assign(matTemp);
      
      // Store Householder vector in the zeroed-out lower-triangle
      // Assuming that the householder vector's first element is equal to one (1)
      // Ignore that first element
      if (iCol < iRows) {
        // A(iCol + 1:iRows, iCol) = v(2:iRows - iCol + 1)
        for (int iOffset = 1; iOffset < iHeight; iOffset++) {
          QR.setQuick(iCol + iOffset, iCol, vecReflector.getQuick(iOffset));
        }
      }
      
//      // System.out.println();
//      // System.out.println("Phase 3 (Upper-triag): " + QR);
      
      
      /**
       * for i = r + 1: n
       *   c(i) = c(i) - A(r, i)^2
       * end
       *
       * A(r,i)^2 contains the square of the component in that vector
       */
      // From the set of remaining column vector norm squares
      // subtract off the norm-square of the current vector
      for (int i = iCol + 1; i < iCols; i++) {
         dAryNormSquared[i] = dAryNormSquared[i] - Math.pow(QR.getQuick(iCol,  i), 2);
      }
//      // System.out.println();
//      // System.out.println("Phase 4 (Norm Squared subtraction): " + QR);

    } // end of column-looping
    
//    // System.out.println();
//    // System.out.println("Last check: ");
//    // System.out.println("Permutations: "); print(iAryIndices);
//    print(dAryNormSquared);
  }
  
  private void print(double[] values) {
    int iLength = values.length;
    // System.out.println("*** Array values ***");
    for (int i = 0; i < iLength; i++) {
      // System.out.println("i: " + values[i]);
    }    
  }
  private void print(int[] values) {
    int iLength = values.length;
    // System.out.println("*** Array values ***");
    for (int i = 0; i < iLength; i++) {
      // System.out.println("i: " + values[i]);
    }    
  }
  
  /**
   * Find the largest norm with the lowest-numbered index (arbitrary but effective)
   */
  private int findLargestValue(double[] values) {
    return findLargestValue(values, 0);
  }
  /**
   * Find the largest norm with the lowest-numbered index (arbitrary but effective)
   * starting from the specified index and on up
   */
  private int findLargestValue(double[] values, int startingIndex) {
    int iSize = values.length;
    // Safety-check
    if (values == null) {throw new IllegalArgumentException("A valid array must be provided");}
    if (iSize == 0) {throw new IllegalArgumentException("A non-empty array must be provided");}
    
    int iMaxIndex = startingIndex;
    double dMax = values[startingIndex];
    for (int i = startingIndex + 1; i < iSize; i++) {
      double dValue = values[i];
      if (dValue > dMax) {
        dMax = dValue;
        iMaxIndex = i;
      }
    }
    
    return iMaxIndex;
  }
  
  
  /**
   * Householder Vector
   * As described by Algorithm 5.1.1 (Householder Vector) p.210
   * x is an element of Real-n
   * v is an element of Real-n; v(1) = 1
   * beta is an element of Real-1
   * P = I_n - beta * v * v'; P is orthogoanl
   * P * x = 2-norm(x) * e_1
   */
  public static class HouseholderVector {
    protected DoubleMatrix1D vecInput = null;
    public HouseholderVector() {}
    public HouseholderVector(DoubleMatrix1D vector) {
      reset(vector);
    }
    
    protected DoubleMatrix1D vecReflection = null;
    public DoubleMatrix1D getReflection() {
      // Safety-check
      if (vecInput == null) {throw new IllegalStateException("The object has not been initialized; use the reset method to initialize.");}

      return vecReflection;
    }
    
    protected double dMultiplier = 0; // aka beta
    public double getMultiplier() {
      // Safety-check
      if (vecInput == null) {throw new IllegalStateException("The object has not been initialized; use the reset method to initialize.");}
      
      return dMultiplier;
    }
    
    protected int n;
    public DoubleMatrix1D reset(DoubleMatrix1D input) {
      // Safety-check
      if (input == null) {throw new IllegalArgumentException("An existing vector must be provided");}
      
      vecInput = input;
      n = vecInput.size();
      
      // Safety-check
      if (n == 0) {throw new IllegalArgumentException("A non-empty vector must be provided.");}
      
      // Safety-check
      if (n == 1) {
        vecReflection = cern.colt.matrix.DoubleFactory1D.dense.make(1, 1);
        dMultiplier = 0;
        return vecReflection;
      }
      
      DoubleMatrix1D vecSubset = vecInput.viewPart(1, n-1);
      double dSigma = vecSubset.zDotProduct(vecSubset);
      
      vecReflection = vecInput.copy();
      vecReflection.setQuick(0, 1); // Set first element to 1
      
      if (dSigma == 0) {
        dMultiplier = 0;
        
      } else {
        double d1stElement = vecInput.getQuick(0);
        double dMu = Math.sqrt(d1stElement*d1stElement + dSigma);
        
        double dReflector1stElement;
        if (d1stElement <= 0) {
          dReflector1stElement = d1stElement - dMu;
        } else {
          dReflector1stElement = -dSigma / (d1stElement + dMu);
        }
        vecReflection.setQuick(0, dReflector1stElement);
                
        dMultiplier = 2 * dReflector1stElement*dReflector1stElement / (dSigma + dReflector1stElement*dReflector1stElement);
        vecReflection.assign(Functions.div(dReflector1stElement));
      }
      
      return vecReflection;
    }

    protected cern.colt.matrix.linalg.Algebra LA = new cern.colt.matrix.linalg.Algebra();
    /**
     * The reflection matrix
     *
     * P = I - [2/(v' * v)] * (v * v')
     */
    public DoubleMatrix2D getReflector() {
      // Safety-check
      if (vecInput == null) {throw new IllegalStateException("The object has not been initialized; use the reset method to initialize.");}
      
      DoubleMatrix2D matP = DoubleFactory2D.dense.identity(n);
      
      double dScalar = dMultiplier;
      DoubleMatrix2D matTemp = LA.multOuter(vecReflection, vecReflection, null);
      matTemp.assign(Functions.mult(dScalar));
      
      matP.assign(matTemp, Functions.minus);
      return matP;
    }
  }
  
  /**
   * Returns the Householder vectors <tt>H</tt>.
   * @return A lower trapezoidal matrix whose columns define the householder reflections.
   */
  public DoubleMatrix2D getH() {    
    DoubleMatrix2D matH = LA.trapezoidalLower(QR.copy());
    
    /** @todo Redo this method (and getQ()) if there are extra (zero-padding) rows 
     for correctness and efficiency*/
    
    // Populate the diagonal with 1 since the first component of the vector has been scaled to 1
    int iDiagonalLength = Math.min(iRows, iCols);
    for (int j = 0; j < iDiagonalLength; j++) {
      matH.setQuick(j, j, 1);
    }
    
    return matH;
  }
  
  /**
   * Get the Q matrix of the QR-decomposition
   */
  public DoubleMatrix2D getQ() {
    return getQ3();
  }

  /**
   * DOES NOT WORK
   * Generates and returns the orthogonal factor <tt>Q</tt>.
   * As described by Algorithm 5.1.2 (Q generation) p.214
   * @return <tt>Q</tt>
   */
  private DoubleMatrix2D getQ1() {
    System.err.println("Method getQ1() does not work!");
    //  DOES NOT WORK
    Functions F = Functions.functions;
    DoubleMatrix2D matHouseholderVectors = getH();
    
    // Y = v_1
    DoubleMatrix2D matY = matHouseholderVectors.viewPart(0, 0, iRows, 1);
    
    // W = -beta * v_1
    double dMultiplier = dAryMultiplier[0];
    DoubleMatrix2D matW = matHouseholderVectors.viewPart(0, 0, iRows, 1).assign( F.mult(-dMultiplier) );
    
    // for j = 2:columns
    for (int iColumn = 1; iColumn < iCols; iColumn++) {
      // v_j
      DoubleMatrix1D vecHouseholder = matHouseholderVectors.viewColumn(iColumn);
      
      // z = -beta * (I + W*Y') * v
      DoubleMatrix2D matTemp = LA.mult(matW, matY.viewDice());
      addIdentity(matTemp);
      DoubleMatrix1D vecZ = LA.mult(matTemp, vecHouseholder);
      dMultiplier = dAryMultiplier[iColumn];
      vecZ = Algebra2.DEFAULT2.multEq(vecZ, -dMultiplier);

      // W = [W z]
      matW = cern.colt.matrix.DoubleFactory2D2.dense.appendColumns(matW, vecZ);
      
      // Y = [Y v_j]
      int iNbrColumns = iColumn + 1;
      matY = matHouseholderVectors.viewPart(0, 0, iRows, iNbrColumns);
    }
    
    
    // Q = I + W * Y'
    DoubleMatrix2D matQ = LA.mult(matW, matY.viewDice());
    addIdentity(matQ);
    
    return matQ;
  }


  /**
   * Generates and returns the orthogonal factor <tt>Q</tt>.
   * As described by Algorithm p.213.
   *
   * @return <tt>Q</tt>
   */
  public DoubleMatrix2D getQ2() {
    DoubleMatrix2D matHouseholderVectors = getH();
    
    DoubleMatrix2D matQ = factory2D.identity(iRows);
    
    // for j = 1:columns
    //for (int iColumn = 0; iColumn < iCols; iColumn++) {
    for (int iColumn = iCols - 1; iColumn >= 0; iColumn--) {
      // v
      DoubleMatrix1D vecHouseholder = matHouseholderVectors.viewColumn(iColumn);
      
      // Qi = P = (I - beta * v * v')      
      DoubleMatrix2D matTemp = LA.multOuter(vecHouseholder, vecHouseholder, null);
      double dScalar = dAryMultiplier[iColumn];      
      matTemp.assign(Functions.mult(dScalar));
      DoubleMatrix2D matP = factory2D.identity(iRows);
      matP.assign(matTemp, Functions.minus);
      
      // Q = Q * Qi for i = 1 to columns, and Q = Qi * Q for i = columns to 1 (because of Q's symmetry)      
      matQ = LA.mult(matP, matQ);
    }

    return matQ;
  }


  /**
   * Same as getQ2 but trims down any extra rows and columns
   */
  public DoubleMatrix2D getQ3() {
    DoubleMatrix2D matHouseholderVectors = getH();
    
    DoubleMatrix2D matQ = factory2D.identity(iRows);
    
    // for j = 1:columns
    for (int iColumn = 0; iColumn < iCols; iColumn++) {
    //for (int iColumn = iCols - 1; iColumn >= 0; iColumn--) {
      // v
      DoubleMatrix1D vecHouseholder = matHouseholderVectors.viewColumn(iColumn);

      // Qi = P = (I - beta * v * v')
      double dScalar = dAryMultiplier[iColumn];
      DoubleMatrix2D matTemp = LA.multOuter(vecHouseholder, vecHouseholder, null);
      matTemp.assign(Functions.mult(dScalar));
      DoubleMatrix2D matP = factory2D.identity(iRows); // TESTING
      matP.assign(matTemp, Functions.minus);      
      
      // Q = Qi * Q
      matQ = LA.mult(matP, matQ);
    }

    // Trim down any extra rows and columns
    matQ = matQ.viewDice(); // Transpose
    if (iExtraRows > 0) {
      int iRowsOriginal = iRows - iExtraRows;
      int[] iAryOriginalRows = new int[iRowsOriginal];
      for (int i = 0; i < iRowsOriginal; i++) {
        iAryOriginalRows[i] = i;        
      }      
      matQ = matQ.viewSelection(iAryOriginalRows, iAryOriginalRows);      
    }
    
    return matQ;
  }

  
  
  protected DoubleMatrix2D addIdentity(DoubleMatrix2D A) {
    int iRows = A.rows();
    int iColumns= A.columns();
    // Safety-check
    if (iRows != iColumns) {throw new IllegalArgumentException("The matrix must be square");}
    
    int iDiagonalLength = iRows;
    for (int j = 0; j < iDiagonalLength; j++) {
      A.setQuick(j,j, A.getQuick(j, j) + 1);
    }
    
    return A;
  }

  
  /**
   * Returns the upper triangular factor, <tt>R</tt>.
   *
   * @return <tt>R</tt>
   */
  public DoubleMatrix2D getR() {
    DoubleMatrix2D matR = Algebra2.DEFAULT2.trapezoidalUpper(QR.copy());
    // Trim down any extra rows
    if (iExtraRows > 0) {
      int iRowsOriginal = iRows - iExtraRows;
      int[] iAryOriginalRows = new int[iRowsOriginal];
      for (int i = 0; i < iRowsOriginal; i++) {
        iAryOriginalRows[i] = i;        
      }      
      matR = matR.viewSelection(iAryOriginalRows, null);
    }
    return matR;
  }

  
  /**
   * Returns whether the matrix <tt>A</tt> has full rank.
   *
   * @return true if <tt>R</tt>, and hence <tt>A</tt>, has full rank.
   */
  public boolean hasFullRank() {
    int iDiagonalLength = Math.min(iRows, iCols);
    for (int j = 0; j < iDiagonalLength; j++) {
      // if (QR.getQuick(j, j) == 0) return false;
      if (Math.abs(QR.getQuick(j, j)) <= dTOL) return false;
    }
    return true;
  }
  

  /** @todo [3] Get Least Squares working */
//  /**
//   * Least squares solution of <tt>A*X = B</tt>; <tt>returns X</tt>.
//   * @param B    A matrix with as many rows as <tt>A</tt> and any number of columns.
//   * @return     <tt>X</tt> that minimizes the two norm of <tt>Q*R*X - B</tt>.
//   * @exception  IllegalArgumentException  if <tt>B.rows() != A.rows()</tt>.
//   * @exception  IllegalArgumentException  if <tt>!this.hasFullRank()</tt> (<tt>A</tt> is rank deficient).
//   */
//  public DoubleMatrix2D solve(DoubleMatrix2D B) {
//    cern.jet.math.Functions F = cern.jet.math.Functions.functions;
//    if (B.rows() != m) {
//      throw new IllegalArgumentException("Matrix row dimensions must agree.");
//    }
//    if (!this.hasFullRank()) {
//      throw new IllegalArgumentException("Matrix is rank deficient.");
//    }
//    
//    // Copy right hand side
//    int nx = B.columns();
//    DoubleMatrix2D X = B.copy();
//    
//    // Compute Y = transpose(Q)*B
//    for (int k = 0; k < n; k++) {
//      for (int j = 0; j < nx; j++) {
//        double s = 0.0;
//        for (int i = k; i < m; i++) {
//          s += QR.getQuick(i,k)*X.getQuick(i,j);
//        }
//        s = -s / QR.getQuick(k,k);
//        for (int i = k; i < m; i++) {
//          X.setQuick(i,j, X.getQuick(i,j) + s*QR.getQuick(i,k));
//        }
//      }
//    }
//    // Solve R*X = Y;
//    for (int k = n-1; k >= 0; k--) {
//      for (int j = 0; j < nx; j++) {
//        X.setQuick(k,j, X.getQuick(k,j) / Rdiag.getQuick(k));
//      }
//      for (int i = 0; i < k; i++) {
//        for (int j = 0; j < nx; j++) {
//          X.setQuick(i,j, X.getQuick(i,j) - X.getQuick(k,j)*QR.getQuick(i,k));
//        }
//      }
//    }
//    return X.viewPart(0,0,n,nx);
//  }

  
  
  private static cern.colt.matrix.linalg.Algebra2 LAStatic = new cern.colt.matrix.linalg.Algebra2();
  private static cern.colt.matrix.DoubleFactory1D factory1DStatic = cern.colt.matrix.DoubleFactory1D.dense;
  /**
   * A is m by n, b is n x 1, m < n
   * p.259
   * 2004-02-25
   */
  public static DoubleMatrix1D solveRankDeficient(DoubleMatrix2D A, DoubleMatrix1D b) {
    // Safety-check
    if (A == null || b == null) {
      throw new IllegalArgumentException("A valid matrix and vector must be provided");      
    }
    int iNbrEquations = A.rows();
    int iNbrVariables = A.columns();
    // Safety-check
    if (b.size() != iNbrEquations) {
      throw new IllegalArgumentException("The number of rows in the matrix must equal the number of elements in the vector");
    }
    if (iNbrEquations == 0) {
      throw new IllegalArgumentException("There must be some equations provided in the matrix");
    }
    if (iNbrVariables == 0) {
      throw new IllegalArgumentException("There must be some variables in the problem");
    }
    
    // Pivoting is critical for this algorithm
    QRDecomposition2 qr = new QRDecomposition2(A, true);
    int[] iAryIndependentColumns = qr.getIndependentColumns();
    int iRank = iAryIndependentColumns.length;

    
    // R = [R11 R12; 0 0]
    DoubleMatrix2D matR = qr.getR();
    DoubleMatrix2D matR11 = matR.viewPart(0, 0, iRank, iRank); // Top-left corner
    
    // Q'*n = [c; d]
    DoubleMatrix2D matQ = qr.getQ();
    DoubleMatrix1D vecQB = LAStatic.mult(matQ.viewDice(), b);
    DoubleMatrix1D vecC = vecQB.viewPart(0, iRank);
    
    // xb = PivotMatrix * (inv(R11) * c; 0)
    DoubleMatrix1D vecTop = LAStatic.solve(matR11, vecC);
    DoubleMatrix1D vecBottom = factory1DStatic.make(iNbrVariables - iRank, 0);
    DoubleMatrix1D vecComplete = factory1DStatic.append(vecTop, vecBottom);

//    System.out.println("Pivots: " + qr.getPivots());
//    System.out.println("vecTop: " + vecTop);
//    System.out.println("vecBottom: " + vecBottom);
//    System.out.println("vecComplete: " + vecComplete);
    DoubleMatrix1D vecX = LAStatic.mult(qr.getPivots(), vecComplete);
    
    return vecX;
  }
  
  
  
  
  
  /**
   * The pivot matrix, E, such that A*E = Q*R
   * where the diagonal elements of R will be decreasing
   * absolute-magnitude order
   */
  public DoubleMatrix2D getPivots() {
    DoubleMatrix2D matPivots = DoubleFactory2D.sparse.make(iCols, iCols);
    for (int iCol = 0; iCol < iCols; iCol++) {
//      // System.out.println("Pivot Col:" + iCol);
//      // System.out.println("Pivot Index Value: " + iAryIndices[iCol]);
      matPivots.setQuick(iAryIndices[iCol], iCol, 1);
    }
    return matPivots;
  }
  
  /**
   * Returns a String with (propertyName, propertyValue) pairs.
   * Useful for debugging or to quickly get the rough picture.
   * For example, 
   * <pre>
   * rank          : 3
   * trace         : 0 
   * </pre>
   */
  public String toString() {
    StringBuffer buf = new StringBuffer();
    String unknown = "Illegal operation or error: ";
    
    buf.append("-----------------------------------------------------------------\n");
    buf.append("QRDecomposition(A) --> hasFullRank(A), H, Q, R, pseudo inverse(A)\n");
    buf.append("-----------------------------------------------------------------\n");
    
    buf.append("hasFullRank = ");
    try { buf.append(String.valueOf(this.hasFullRank()));}
    catch (IllegalArgumentException exc) { buf.append(unknown+exc.getMessage()); }
    
//    buf.append("\n\nH = ");
//    try { buf.append(String.valueOf(this.getH()));}
//    catch (IllegalArgumentException exc) { buf.append(unknown+exc.getMessage()); }
    
    buf.append("\n\nQ = ");
    try { buf.append(String.valueOf(this.getQ()));}
    catch (IllegalArgumentException exc) { buf.append(unknown+exc.getMessage()); }
    
    buf.append("\n\nR = ");
    try { buf.append(String.valueOf(this.getR()));}
    catch (IllegalArgumentException exc) { buf.append(unknown+exc.getMessage()); }
    
//    buf.append("\n\npseudo inverse(A) = ");
//    try { buf.append(String.valueOf(this.solve(cern.colt.matrix.factory2D.identity(QR.rows()))));}
//    catch (IllegalArgumentException exc) { buf.append(unknown+exc.getMessage()); }
    
    return buf.toString();
  }

  /**
   * Swap the values of two elements within an array of integer primitives
   */
  protected void swap(int[] values, int firstIndex, int secondIndex) {
    // Safety-check
    if (values == null) {throw new IllegalArgumentException("An array of values must be provided.");}
    int iLength = values.length;
    if (firstIndex < 0 || secondIndex < 0 || firstIndex >= iLength || secondIndex >= iLength) {
      throw new IllegalArgumentException("An invalid index was specified for the swap");      
    }
    
    int iTemp = values[firstIndex];
    values[firstIndex] = values[secondIndex];
    values[secondIndex] = iTemp;
  }
  
  /**
   * Swap the values of two elements within an array of double primitives
   */
  protected void swap(double[] values, int firstIndex, int secondIndex) {
    // Safety-check
    if (values == null) {throw new IllegalArgumentException("An array of values must be provided.");}
    int iLength = values.length;
    if (firstIndex < 0 || secondIndex < 0 || firstIndex >= iLength || secondIndex >= iLength) {
      throw new IllegalArgumentException("An invalid index was specified for the swap");      
    }
    
    double iTemp = values[firstIndex];
    values[firstIndex] = values[secondIndex];
    values[secondIndex] = iTemp;
  }  
  
  /**
   * Determine which columns of the original matrix are redundant
   * These columns cause the matrix to be singular
   * Since the redundant columns are linear combinations of at least
   * one other column in the matrix, the identification of these columns
   * is not unique (ie arbitrary).  The norm of the "non-redundant" columns are >=
   * the norm of the "redundant" columns.
   *
    * @return An array of column indices containing the redundant columns of the original matrix
   */
  public int[] getDependentColumns() {
    IntegerList list = new IntegerList();
    for (int iCol = 0; iCol < iCols; iCol++) {
      double dColumnMagnitude = Math.abs(QR.getQuick(iCol, iCol));
      // System.out.println("QR iCol: " + iCol + "; value = " + dColumnMagnitude);
      if (dColumnMagnitude <= dTOL || Double.isNaN(dColumnMagnitude)) {
        // Dependent columns
        list.add(iAryIndices[iCol]);
      }
    }
    
    int[] iAryIndex = list.toArray();
    // java.util.Arrays.sort(iAryIndex);
    return iAryIndex;
  }
  
  /**
   * Determine which columns of the original matrix are independent/not redundant.
   *
   * @return An array of column indices containing the independent columns of the original matrix.
   */
  public int[] getIndependentColumns() {
    IntegerList list = new IntegerList();
    for (int iCol = 0; iCol < iCols; iCol++) {
      double dColumnMagnitude = Math.abs(QR.getQuick(iCol, iCol));
      if (dColumnMagnitude > dTOL) {
        // Independent columns
        list.add(iAryIndices[iCol]);
      }
    }
    
    int[] iAryIndex = list.toArray();    
    return iAryIndex;
  }

  protected double dTOL = 1e-6;
  /**
   * Get the tolerance value used  in identifying dependent columns, independent columns,
   * and determining whether a matrix has full rank.
   */
  public double getTolerance() {return dTOL;}
  /**
   * Set the tolerance value used in identifying dependent columns, independent columns,
   * and determining whether a matrix has full rank.
   */
  public void setTolerance(double tolerance) {
    // Safety-check
    if (tolerance <= 0) {
      throw new IllegalArgumentException("Tolerance must be a positive");      
    }
    
    this.dTOL = tolerance;
  }
  
}
