/*
 * Copyright (C) 2003 Michel Santos
 * Licensed under the Academic Free License version 3.0
 * Please see the LICENSE for more information.
 *
 * Created on April 12, 2003, 3:27 PM
 */

package org.xspace.util;

/**
 * A container for a variable number of integer primitives.
 *
 * @author  Michel Santos
 */
public class IntegerList {
  
  /** 
   * Creates a new instance of IntegerList 
   */
  public IntegerList() {
    init(10);
  }
  
  /**
   * Creates a new instance of IntegerList 
   */
  public IntegerList(int initialCapacity)  {
    init(initialCapacity);
  }
  
  /**
   * Creates a new instance of IntegerList 
   */
  public IntegerList(IntegerList other)  {
    int iSize = other.getSize();
    init(iSize);
    add(other.toArray());
  } 
  
  /**
   * Creates a new instance of IntegerList 
   */
  public IntegerList(int[] values)  {
    int iSize = values.length;
    init(iSize);
    add(values);
  } 

  
  protected void init(int initialCapacity)  {
    // Safety-check
    if (initialCapacity < 1) {
      initialCapacity = 1;
    }
    
    iCapacity = initialCapacity;
    iArray = new int[iCapacity];
  }
  
  
  protected int iCapacity;
  protected int iCurrentSize = 0;
  /**
   * Get the size of the container
   */
  public int getSize() {return iCurrentSize;}
  protected int[] iArray;

  /**
   * Add an integer to the container
   */
  public void add(int value) {
    // Safety-check
    if (iCurrentSize + 1 > iCapacity) {
      // Increase size of internal storage
      int iNewSize = 2 * iCapacity;
      int[] iArrayNew = new int[iNewSize];
      System.arraycopy(iArray, 0, iArrayNew, 0, iCapacity);
      iArray = iArrayNew;
      iCapacity = iNewSize;
    }
    
    iArray[iCurrentSize] = value;
    iCurrentSize++;
  }
  
  /**
   * Add an array of integers to the container
   */
  public void add(int[] values) {
    int iBatchSize = values.length;

    // Safety-check
    if (iBatchSize == 0) {
      return;      
    }    
    
    
    // Safety-check
    if (iCurrentSize + iBatchSize > iCapacity) {
      // Increase size of internal storage      
      int iNewSize = Math.max(2 * iCapacity, iCapacity + iBatchSize);
      int[] iArrayNew = new int[iNewSize];
      System.arraycopy(iArray, 0, iArrayNew, 0, iCapacity);
      iArray = iArrayNew;
      iCapacity = iNewSize;
    }

    
    System.arraycopy(values, 0, iArray, iCurrentSize, iBatchSize);
    iCurrentSize += iBatchSize;
  }
  
  /**
   * Obtain an integer array with the contents of this container
   */
  public int[] toArray() {
    // Return the subset of the internal storage that has been "add"ed
    int[] iArraySubset = new int[iCurrentSize];
    System.arraycopy(iArray, 0, iArraySubset, 0, iCurrentSize);    
    return iArraySubset;
  }  
  
  /**
   * Get the integer located at a particular index.  The container is zero-based.
   *
   * @parameter index Index of the integer being requested
   * @return  The integer primitive requested
   */
  public int get(int index) {
    // Safety-check
    if (index < 0) {throw new IndexOutOfBoundsException("The requested index was negative.");}
    if (index >= iCurrentSize) {throw new IndexOutOfBoundsException("The requested index was beyond the range of valid indices.");}
    
    return iArray[index];    
  }
  
  /**
   * Clears the container of its contents
   */
  public void clear() {
    java.util.Arrays.fill(iArray, 0);
    iCurrentSize = 0;
  }
  
  
  /**
   * Delete the integer located at a particular index.  The container is zero-based.
   *
   * @parameter index Index of the integer being deleted
   */
  public void delete(int index) {
    // Safety-check
    if ( (index < 0) || (index > (iCurrentSize - 1)) ) {
      throw new IndexOutOfBoundsException("Attempted to delete slot " + index + " from a list of only " + iCurrentSize + " items");
    }
    
    // Shift everything above deletion slot down
    if ((index + 1) == iCapacity) {
      // If the slot being deleted is the last slot available (iCapacity)
      // then there is nothing to "shift down"
      // Do nothing
    } else {
      // Shift the "higher" contents "down"
      int iNbrSlotsToTransfer = iCurrentSize - (index + 1);
      System.arraycopy(iArray, index + 1, iArray, index, iNbrSlotsToTransfer);
    }
    iCurrentSize--;
  }  

  public String toString() {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < iCurrentSize; i++) {
      sb.append(iArray[i]).append(" ");
    }
    return sb.toString();
  }
  
}